
/* Chrysalide - Outil d'analyse de fichiers binaires
 * area.c - définition et manipulation des aires à désassembler
 *
 * Copyright (C) 2014-2017 Cyrille Bagard
 *
 *  This file is part of Chrysalide.
 *
 *  Chrysalide is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Chrysalide is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "area.h"


#include <assert.h>
#include <malloc.h>
#include <string.h>


#include <i18n.h>


#include "../../analysis/contents/restricted.h"
#include "../../arch/raw.h"
#include "../../common/bits.h"
#include "../../common/sort.h"
#include "../../format/format.h"
#include "../../glibext/delayed-int.h"
#include "../../gui/panels/log.h"



/* ------------------------- TRAITEMENT DES ZONES DE DONNES ------------------------- */


/* Zone mémoire bien bornée */
typedef struct _mem_area
{
    GBinFormat *format;                     /* Format du fichier binaire   */
    GBinContent *content;                   /* Données binaires à lire     */
    GArchProcessor *proc;                   /* Architecture du binaire     */
    SourceEndian endianness;                /* Boutisme de cette machine   */

    mrange_t range;                         /* Couverture de la zone       */

    phys_t packing_size;                    /* Granularité des découpages  */

    bitfield_t *processed;                  /* Octets traités dans la zone */
    GArchInstruction **instructions;        /* Instructions en place       */
    size_t count;                           /* Quantité d'instructions     */

    bool is_exec;                           /* Zone exécutable ?           */

} mem_area;


/* Initialise une aire de données à partir d'une adresse donnée. */
static void init_mem_area_from_addr(mem_area *, const vmpa2t *, phys_t, const GLoadedBinary *);

/* Libère d'une aire de données les ressources allouées. */
static void fini_mem_area(mem_area *);

/* Indique si une zone donnée est intégralement vierge ou non. */
static bool is_range_blank_in_mem_area(mem_area *, phys_t, phys_t);

/* Marque une série d'octets comme ayant été traités. */
static bool mark_range_in_mem_area_as_processed(mem_area *, GArchInstruction *, bool);

/* Crée une instruction issue d'un désassemblage brut. */
static GArchInstruction *load_raw_instruction_from_mem_area(mem_area *, phys_t, vmpa2t *, phys_t *);

/* S'assure de la présence d'un début de routine à un point. */
static void update_address_as_routine(GBinFormat *, const vmpa2t *);

/* Procède au désassemblage d'un contenu binaire non exécutable. */
static void load_data_from_mem_area(mem_area *, GProcContext *, const vmpa2t *, GtkStatusStack *, activity_id_t);

/* S'assure qu'une aire contient toutes ses instructions. */
static void fill_mem_area(mem_area *, mem_area *, size_t, GProcContext *, GtkStatusStack *, activity_id_t);

/* Rassemble les instructions conservées dans une zone donnée. */
static GArchInstruction **get_instructions_from_mem_area(const mem_area *, GArchInstruction **, size_t *);



/* ----------------------- MANIPULATIONS PARALLELES DES ZONES ----------------------- */


#define G_TYPE_AREA_COLLECTOR            g_area_collector_get_type()
#define G_AREA_COLLECTOR(obj)            (G_TYPE_CHECK_INSTANCE_CAST((obj), G_TYPE_AREA_COLLECTOR, GAreaCollector))
#define G_IS_AREA_COLLECTOR(obj)         (G_TYPE_CHECK_INSTANCE_TYPE((obj), G_TYPE_AREA_COLLECTOR))
#define G_AREA_COLLECTOR_CLASS(klass)    (G_TYPE_CHECK_CLASS_CAST((klass), G_TYPE_AREA_COLLECTOR, GAreaCollectorClass))
#define G_IS_AREA_COLLECTOR_CLASS(klass) (G_TYPE_CHECK_CLASS_TYPE((klass), G_TYPE_AREA_COLLECTOR))
#define G_AREA_COLLECTOR_GET_CLASS(obj)  (G_TYPE_INSTANCE_GET_CLASS((obj), G_TYPE_AREA_COLLECTOR, GAreaCollectorClass))


/* Ensembles binaires à désassembler (instance) */
typedef struct _GAreaCollector
{
    GDelayedWork parent;                    /* A laisser en premier        */

    activity_id_t id;                       /* Groupe de progression       */
    run_task_fc run;                        /* Activité dans la pratique   */

    mem_area *areas;                        /* Zone de productions         */

    union
    {
        struct
        {
            size_t acount;                  /* Nombre de zones créées      */

            GLoadedBinary *binary;          /* Binaire à associer aux zones*/

            phys_t first;                   /* Début de traitement         */
            phys_t last;                    /* Fin de traitement           */

            bool closing;                   /* Tâche clôturant le parcours */

        };

        struct
        {
            size_t begin;                   /* Début du parcours à mener   */
            size_t end;                     /* Fin de ce même parcours     */

            GArchInstruction **collected;   /* Instructions collectées     */
            size_t ccount;                  /* Quantité de ces instructions*/

        };

    };

} GAreaCollector;

/* Ensembles binaires à désassembler (classe) */
typedef struct _GAreaCollectorClass
{
    GDelayedWorkClass parent;               /* A laisser en premier        */

} GAreaCollectorClass;


/* Indique le type défini pour les tâches de traitement des zones. */
GType g_area_collector_get_type(void);

/* Initialise la classe des manipulations parallèles de zones. */
static void g_area_collector_class_init(GAreaCollectorClass *);

/* Initialise des manipulations parallèles de zones. */
static void g_area_collector_init(GAreaCollector *);

/* Supprime toutes les références externes. */
static void g_area_collector_dispose(GAreaCollector *);

/* Procède à la libération totale de la mémoire. */
static void g_area_collector_finalize(GAreaCollector *);

/* Assure un traitement particulier concernant les zones. */
static void g_area_collector_process(GAreaCollector *, GtkStatusStack *);

/* Crée une tâche de calcul des zones binaires à désassembler. */
static GAreaCollector *g_area_collector_new_intro(activity_id_t, GLoadedBinary *, phys_t, phys_t, bool);

/* Construit une liste bornée de zones contigües. */
static void g_area_collector_do_compute(GAreaCollector *, GtkStatusStack *);

/* Crée une tâche de récupération d'instructions différée. */
static GAreaCollector *g_area_collector_new_outro(activity_id_t, mem_area *, size_t, size_t);

/* Assure la récupération d'instructions en différé. */
static void g_area_collector_do_collect(GAreaCollector *, GtkStatusStack *);



/* ---------------------------------------------------------------------------------- */
/*                            RAITEMENT DES ZONES DE DONNES                           */
/* ---------------------------------------------------------------------------------- */


/******************************************************************************
*                                                                             *
*  Paramètres  : area   = aire représentant à contenu à initialiser.          *
*                addr   = adresse de départ de l'espace à mettre en place.    *
*                len    = longueur de l'espace à créer.                       *
*                binary = binaire analysé content quantités d'informations.   *
*                                                                             *
*  Description : Initialise une aire de données à partir d'une adresse donnée.*
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void init_mem_area_from_addr(mem_area *area, const vmpa2t *addr, phys_t len, const GLoadedBinary *binary)
{
    GBinContent *content;                   /* Données binaires à lire     */

    assert(len > 0);

    area->format = G_BIN_FORMAT(g_loaded_binary_get_format(binary));

    content = g_binary_format_get_content(area->format);

    area->proc = g_loaded_binary_get_processor(binary);
    area->endianness = g_arch_processor_get_endianness(area->proc);

    init_mrange(&area->range, addr, len);

    area->content = g_restricted_content_new(content, &area->range);

    area->packing_size = 2; /* FIXME */

    area->processed = create_bit_field(len, false);
    area->instructions = (GArchInstruction **)calloc(len, sizeof(GArchInstruction *));
    area->count = 0;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : area = aire représentant à contenu à nettoyer en mémoire.    *
*                                                                             *
*  Description : Libère d'une aire de données les ressources allouées.        *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void fini_mem_area(mem_area *area)
{
    phys_t len;                             /* Etendue du parcours total   */
    phys_t i;                               /* Boucle de parcours          */

    g_object_unref(area->format);
    g_object_unref(area->content);
    g_object_unref(area->proc);

    delete_bit_field(area->processed);

    len = get_mrange_length(&area->range);

    for (i = 0; i < len; i++)
        if (area->instructions[i] != NULL)
            g_object_unref(G_OBJECT(area->instructions[i]));

    free(area->instructions);

}


/******************************************************************************
*                                                                             *
*  Paramètres  : area  = aire représentant à contenu à parcourir.             *
*                start = début de la zone à manipuler.                        *
*                len   = taille de cette même aire de données.                *
*                                                                             *
*  Description : Indique si une zone donnée est intégralement vierge ou non.  *
*                                                                             *
*  Retour      : true si l'aire visée n'a jamais été traitée, false sinon.    *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static bool is_range_blank_in_mem_area(mem_area *area, phys_t start, phys_t len)
{
    bool result;                            /* Résultat à renvoyer         */

    assert((start + len) <= get_mrange_length(&area->range));

    result = !test_in_bit_field(area->processed, start, len);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : area  = aire représentant à contenu à parcourir.             *
*                instr = instruction à mémoriser pour la suite.               *
*                force = impose l'enregistrement de l'instruction.            *
*                                                                             *
*  Description : Marque une série d'octets comme ayant été traités.           *
*                                                                             *
*  Retour      : true si l'enregistrement a bien été réalisé, false sinon.    *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static bool mark_range_in_mem_area_as_processed(mem_area *area, GArchInstruction *instr, bool force)
{
    bool result;                            /* Bilan d'action à renvoyer   */
    const vmpa2t *start;                    /* Adresse de départ de la zone*/
    const mrange_t *range;                  /* Emplacement d'instruction   */
    const vmpa2t *addr;                     /* Début de la zone à traiter  */
    phys_t len;                             /* Taille de l'aire visée      */
    phys_t offset;                          /* Décallage de départ         */
    phys_t i;                               /* Boucle de parcours          */
    GArchInstruction *old;                  /* Instruction remplacée       */

    start = get_mrange_addr(&area->range);

    range = g_arch_instruction_get_range(instr);
    addr = get_mrange_addr(range);
    len = get_mrange_length(range);

    offset = compute_vmpa_diff(start, addr);

    result = set_atomic_in_bit_field(area->processed, offset, len);

    /* Si l'instruction était bien la première à s'inscrire... */

    result |= force;

    if (result)
    {
        assert(area->instructions[offset] == NULL || force);

        /**
         * Un cas de remplacement forcé intervient en ARM, lorsque qu'une
         * instruction utilise une valeur immédiate placée dans le code.
         *
         * Cette valeur est référencée en tant que symbole.
         *
         * Mais cette même valeur a pu être désassemblée en tant que code
         * exécutable si le flot d'exécution s'est poursuivi jusqu'à elle.
         *
         * C'est par exemple le cas lors de l'utilisation d'appels système
         * en assembleur, qui ne sont pas reconnus en tant qu'instructions
         * cassant le flot d'exécution (typiquement : un exit()).
         *
         * On réinitialise donc la zone couverte par la nouvelle instruction.
         */

        for (i = 0; force && i < len; i++)
        {
            old = area->instructions[offset + i];

            if (old != NULL)
            {
                g_object_unref(G_OBJECT(old));
                area->instructions[offset + 1] = NULL;
                g_atomic_pointer_add(&area->count, -1);
            }

        }

        area->instructions[offset] = instr;
        g_atomic_pointer_add(&area->count, 1);

        /* Au passage, association du contenu */
        g_arch_instruction_set_global_content(instr, area->content);

    }

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : area   = aire représentant à contenu à parcourir.            *
*                offset = point de départ au sein de l'aire en question.      *
*                pos    = tête de lecture dans l'espace global.               *
*                size   = taille de l'instruction mise en place. [OUT]        *
*                                                                             *
*  Description : Crée une instruction issue d'un désassemblage brut.          *
*                                                                             *
*  Retour      : Instruction mise en place ou NULL en cas d'échec.            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static GArchInstruction *load_raw_instruction_from_mem_area(mem_area *area, phys_t offset, vmpa2t *pos, phys_t *size)
{
    GArchInstruction *result;               /* Instruction à retourner     */
    GBinContent *content;                   /* Données binaires à lire     */
    SourceEndian endianness;                /* Boutisme de cette machine   */
    phys_t sz;                              /* Volume de données traité    */
    vmpa2t prev;                            /* Boucle de parcours          */

    result = NULL;

    content = area->content;
    endianness = area->endianness;

    sz = area->packing_size;

    if (get_virt_addr(pos) % sz == 0 && is_range_blank_in_mem_area(area, offset, sz))
    {
        *size = sz;

        copy_vmpa(&prev, pos);

        result = g_raw_instruction_new_array(content, MDS_FROM_BYTES(sz), 1, pos, endianness);

        if (result == NULL)
            copy_vmpa(pos, &prev);

    }

    if (result == NULL)
    {
        *size = 1;

        result = g_raw_instruction_new_array(content, MDS_8_BITS, 1, pos, endianness);

    }

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : format = format binaire en cours de traitement.              *
*                addr   = adresse d'une instruction présentée comme première. *
*                                                                             *
*  Description : S'assure de la présence d'un début de routine à un point.    *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void update_address_as_routine(GBinFormat *format, const vmpa2t *addr)
{
    GBinSymbol *symbol;                     /* Symbole présent ou créé     */
    bool found;                             /* Détection de symbole        */
    SymbolType sym_type;                    /* Type de symbole en place    */
    bool wrong_type;                        /* Analyse plus fine de ce type*/
    mrange_t range;                         /* Etendue du symbole à créer  */
    VMPA_BUFFER(loc);                       /* Traduction de l'adresse     */
    char name[5 + VMPA_MAX_LEN];            /* Nom de symbole nouveau      */
    GBinRoutine *routine;                   /* Nouvelle routine trouvée    */

    found = g_binary_format_find_symbol_at(format, addr, &symbol);

    if (found)
    {
        sym_type = g_binary_symbol_get_target_type(symbol);
        wrong_type = (sym_type != STP_ROUTINE && sym_type != STP_ENTRY_POINT);
    }

    if (!found || (found && wrong_type))
    {
        init_mrange(&range, addr, 0);

        vmpa2_virt_to_string(addr, MDS_UNDEFINED, loc, NULL);
        snprintf(name, sizeof(name), "sub_%s", loc + 2);

        routine = g_binary_routine_new();
        g_binary_routine_set_name(routine, strdup(name));

        g_binary_routine_set_range(routine, &range);

        if (!found)
        {
            symbol = g_binary_symbol_new(STP_ROUTINE);
            g_binary_symbol_attach_routine(symbol, routine);
            g_binary_format_add_symbol(format, symbol);
        }
        else _g_binary_symbol_attach_routine(symbol, routine, STP_ROUTINE);

    }

    if (found)
        g_object_unref(G_OBJECT(symbol));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : area   = aire représentant à contenu à parcourir.            *
*                list   = liste de zones délimitant des contenus à traiter.   *
*                count  = nombre de zones à disposition.                      *
*                index  = indice de l'aire à considérer pendant l'opération.  *
*                binary = représentation de binaire chargé.                   *
*                ctx    = contexte offert en soutien à un désassemblage.      *
*                start  = démarrage de l'exécution au sein de la zone.        *
*                force  = force la création d'au moins une instruction.       *
*                status = barre de statut à actualiser.                       *
*                id     = identifiant du groupe de progression à l'affichage. *
*                                                                             *
*  Description : Procède au désassemblage d'un contenu binaire exécutable.    *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

void load_code_from_mem_area(mem_area *area, mem_area *list, size_t count, GProcContext *ctx, const vmpa2t *start, bool force, GtkStatusStack *status, activity_id_t id)
{



    GBinFormat *format;                     /* Format du fichier binaire   */
    GArchProcessor *proc;                   /* Architecture du binaire     */
    GBinContent *content;                   /* Données binaires à lire     */

    phys_t init_diff;                       /* Position initiale de lecture*/
    phys_t alen;                            /* Taille de l'aire utilisée   */

    bool forced_once;                       /* Préfigure une sortie rapide */

    phys_t i;                               /* Boucle de parcours          */


    vmpa2t pos;                             /* Boucle de parcours          */
    vmpa2t prev;                            /* Boucle de parcours          */

    GArchInstruction *instr;                /* Instruction décodée         */
    phys_t diff;                            /* Volume de données traité    */



    mrange_t range;                         /* Couverture de l'instruction */


    bool done;                              /* Enregistrement effectué ?   */


    vmpa2t sym_addr;                        /* Adresse de nouveau symbole  */
    bool has_new_sym;                       /* Statut d'un dépilement      */

    GBinSymbol *symbol;                     /* Symbole créé en parallèle   */



    /* Récupération des informations de base */

    format = area->format;
    proc = area->proc;
    content = area->content;

    init_diff = compute_vmpa_diff(get_mrange_addr(&area->range), start);
    alen = get_mrange_length(&area->range);

    copy_vmpa(&pos, start);

    /* Traitement de la zone */

    forced_once = false;

    for (i = init_diff; i < alen; i += diff)
    {
        /**
         * On réalise un premier test informel (car non atomique) peu coûteux
         * avant de se lancer dans un désassemblage d'instruction potentiellement
         * inutile.
         */

        if (!is_range_blank_in_mem_area(area, i, 1))
            break;

        /* Décodage d'une nouvelle instruction */

        copy_vmpa(&prev, &pos);

        instr = g_arch_processor_disassemble(proc, ctx, content, &pos, G_EXE_FORMAT(format));

        if (instr != NULL)
            diff = compute_vmpa_diff(&prev, &pos);

        else
        {
            if (i == init_diff && force)
            {
                instr = load_raw_instruction_from_mem_area(area, i, &pos, &diff);
                forced_once = true;
            }

            if (instr == NULL)
                break;

        }

        /* Enregistrement des positions et adresses */

        init_mrange(&range, &prev, diff);

        g_arch_instruction_set_range(instr, &range);

        /* Progression dans les traitements */

        done = mark_range_in_mem_area_as_processed(area, instr, false);

        if (!done)
        {
            g_object_unref(G_OBJECT(instr));
            break;
        }

        gtk_status_stack_update_activity_value(status, id, diff);

        assert(!is_range_blank_in_mem_area(area, i, diff));

        /* Enregistrement d'un éventuel début de routine */

        if (g_arch_instruction_get_flags(instr) & AIF_ROUTINE_START)
            update_address_as_routine(format, &prev);

        /* Eventuel renvoi vers d'autres adresses */

        g_arch_instruction_call_hook(instr, IPH_FETCH, proc, ctx, G_EXE_FORMAT(format));

        /* Insertion des symboles découverts en parallèle */

        for (has_new_sym = g_proc_context_pop_new_symbol_at(ctx, &sym_addr);
             has_new_sym;
             has_new_sym = g_proc_context_pop_new_symbol_at(ctx, &sym_addr))
        {
            has_new_sym = g_binary_format_find_symbol_at(format, &sym_addr, &symbol);

            if (has_new_sym)
            {
                insert_extra_symbol_into_mem_areas(list, count, symbol);
                g_object_unref(G_OBJECT(symbol));
            }

        }

        /* Rupture du flot d'exécution ? */
        if (forced_once || g_arch_instruction_get_flags(instr) & AIF_RETURN_POINT)
            break;

    }

}


/******************************************************************************
*                                                                             *
*  Paramètres  : area   = aire représentant à contenu à parcourir.            *
*                ctx    = contexte offert en soutien à un désassemblage.      *
*                start  = démarrage de l'exécution au sein de la zone.        *
*                status = barre de statut à actualiser.                       *
*                id     = identifiant du groupe de progression à l'affichage. *
*                                                                             *
*  Description : Procède au désassemblage d'un contenu binaire non exécutable.*
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void load_data_from_mem_area(mem_area *area, GProcContext *ctx, const vmpa2t *start, GtkStatusStack *status, activity_id_t id)
{
    phys_t diff;                            /* Volume de données traité    */
    phys_t alen;                            /* Taille de l'aire utilisée   */
    vmpa2t pos;                             /* Boucle de parcours          */
    phys_t i;                               /* Boucle de parcours          */
    vmpa2t prev;                            /* Boucle de parcours          */
    GArchInstruction *instr;                /* Instruction décodée         */
    mrange_t range;                         /* Couverture de l'instruction */
    bool done;                              /* Enregistrement effectué ?   */

    /* Récupération des informations de base */

    diff = compute_vmpa_diff(get_mrange_addr(&area->range), start);
    alen = get_mrange_length(&area->range);

    copy_vmpa(&pos, start);

    /* Traitement de la zone */

    for (i = diff; i < alen; i += diff)
    {
        /* On cherche à obtenir l'assurance que le traitement n'a jamais été fait */

        if (!is_range_blank_in_mem_area(area, i, 1))
            break;

        instr = NULL;

        copy_vmpa(&prev, &pos);

        /* Décodage d'une nouvelle instruction, sur mesure puis minimale */

        instr = load_raw_instruction_from_mem_area(area, i, &pos, &diff);

        /* On rencontre ici un morceau déjà traité. */

        if (instr == NULL) break;

        /* Enregistrement des positions et adresses */

        assert(diff == compute_vmpa_diff(&prev, &pos));

        init_mrange(&range, &prev, diff);

        g_arch_instruction_set_range(instr, &range);

        /* Progression dans les traitements */

        done = mark_range_in_mem_area_as_processed(area, instr, false);

        if (!done)
        {
            g_object_unref(G_OBJECT(instr));
            break;
        }

        gtk_status_stack_update_activity_value(status, id, diff);

        assert(!is_range_blank_in_mem_area(area, i, diff));

        /* On laisse une chance au code pour se reprendre... */

        if (area->is_exec) break;

    }

}


/******************************************************************************
*                                                                             *
*  Paramètres  : area   = aire représentant à contenu à parcourir.            *
*                list   = liste de zones délimitant des contenus à traiter.   *
*                count  = nombre de zones à disposition.                      *
*                binary = représentation de binaire chargé.                   *
*                ctx    = contexte offert en soutien à un désassemblage.      *
*                status = barre de statut à actualiser.                       *
*                id     = identifiant du groupe de progression à l'affichage. *
*                                                                             *
*  Description : S'assure qu'une aire contient toutes ses instructions.       *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void fill_mem_area(mem_area *area, mem_area *list, size_t count, GProcContext *ctx, GtkStatusStack *status, activity_id_t id)
{
    const vmpa2t *addr;                     /* Début de la zone à traiter  */
    phys_t len;                             /* Taille de la zone à remplir */
    phys_t i;                               /* Boucle de parcours          */
    vmpa2t start;                           /* Adresse de départ de combles*/

    addr = get_mrange_addr(&area->range);
    len = get_mrange_length(&area->range);

    for (i = 0; i < len; i++)
    {
        if (is_range_blank_in_mem_area(area, i, 1))
        {
            copy_vmpa(&start, addr);
            advance_vmpa(&start, i);

            if (area->is_exec && get_virt_addr(&start) % area->packing_size == 0)
                load_code_from_mem_area(area, list, count, ctx, &start, false, status, id);

            if (is_range_blank_in_mem_area(area, i, 1))
                load_data_from_mem_area(area, ctx, &start, status, id);

        }

        assert(!is_range_blank_in_mem_area(area, i, 1));

    }

}


/******************************************************************************
*                                                                             *
*  Paramètres  : area  = aire représentant à contenu à parcourir.             *
*                list  = liste d'instructions à compléter.                    *
*                count = taille de cette liste. [OUT]                         *
*                                                                             *
*  Description : Rassemble les instructions conservées dans une zone donnée.  *
*                                                                             *
*  Retour      : Liste d'instructions prêtes à emploi.                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static GArchInstruction **get_instructions_from_mem_area(const mem_area *area, GArchInstruction **list, size_t *count)
{
    GArchInstruction **result;              /* Liste d'instr. à renvoyer   */
    phys_t len;                             /* Nombre d'instructions au max*/
#ifndef NDEBUG
    size_t check;                           /* Verification de débordement */
#endif
    phys_t i;                               /* Boucle de parcours          */
    GArchInstruction *instr;                /* Instruction décodée         */

    result = (GArchInstruction **)realloc(list, (*count + area->count) * sizeof(GArchInstruction *));

    len = get_mrange_length(&area->range);

#ifndef NDEBUG
    check = 0;
#endif

    for (i = 0; i < len; i++)
    {
        instr = area->instructions[i];

        if (instr != NULL)
        {
            g_object_ref(G_OBJECT(instr));
            result[(*count)++] = instr;

#ifndef NDEBUG
            check++;
            assert(check <= area->count);
#endif

        }

    }

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : list  = listes de zones utable à consulter.                  *
*                count = nombre de zones mises en place.                      *
*                addr  = adresse à retrouver dans les aires présentes.        *
*                                                                             *
*  Description : Détermine une liste de zones contigües à traiter.            *
*                                                                             *
*  Retour      : Indice de la zone trouvée, ou nombre d'aires en cas d'échec. *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

mem_area *find_memory_area_by_addr(mem_area *list, size_t count, const vmpa2t *addr)
{
    mem_area *result;                       /* Elément trouvé à renvoyer   */

    int find_mem_area(const vmpa2t *_addr, const mem_area *_area)
    {
        int status;                         /* Bilan à retourner           */

        if (mrange_contains_addr(&_area->range, _addr))
            status = 0;

        else
            status = cmp_vmpa(_addr, get_mrange_addr(&_area->range));

        return status;

    }

    result = bsearch(addr, list, count, sizeof(mem_area), (__compar_fn_t)find_mem_area);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : areas  = liste de zones délimitant des contenus à traiter.   *
*                count  = nombre de zones à disposition.                      *
*                symbol = élément nouveau à venir insérer dans les zones.     *
*                                                                             *
*  Description : Insère un symbole dans un découpage en aires.                *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

void insert_extra_symbol_into_mem_areas(mem_area *areas, size_t count, const GBinSymbol *symbol)
{
    SymbolType type;                        /* Type de symbole             */
    GArchInstruction *instr;                /* Instruction à insérer       */
    const mrange_t *range;                  /* Emplacement d'instruction   */
    const vmpa2t *addr;                     /* Départ de cet emplacement   */
    mem_area *area;                         /* Zone d'accueil désignée     */
    VMPA_BUFFER(loc);                       /* Description d'un emplacement*/
    phys_t start;                           /* Point de départ             */

    type = g_binary_symbol_get_target_type(symbol);

    if (!HAS_DATA_INSTR(type))
        return;

    instr = g_binary_symbol_get_instruction(symbol);

    range = g_arch_instruction_get_range(instr);
    addr = get_mrange_addr(range);

    /* Une aire d'accueil existe-t-elle ? */

    area = find_memory_area_by_addr(areas, count, addr);

    if (area == NULL)
    {
        vmpa2_virt_to_string(addr, MDS_UNDEFINED, loc, NULL);

        log_variadic_message(LMT_WARNING, _("No place found for symbol located at %s."), loc);
        return;

    }

    /* L'instruction est-elle accueillie dans son intégralité ? */

    start = compute_vmpa_diff(get_mrange_addr(&area->range), addr);

    if (start + get_mrange_length(range) > get_mrange_length(&area->range))
    {
        vmpa2_virt_to_string(addr, MDS_UNDEFINED, loc, NULL);

        log_variadic_message(LMT_WARNING, _("The symbol located at %s is too big for one place only."), loc);
        return;

    }

    /* Inscription d'une instruction de symbole (sans retour arrière possible :/ ) */

    mark_range_in_mem_area_as_processed(area, instr, true);

    g_object_ref(G_OBJECT(instr));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : list   = liste de zones délimitant des contenus à traiter.   *
*                count  = nombre de zones à disposition.                      *
*                binary = représentation de binaire chargé.                   *
*                ctx    = contexte offert en soutien à un désassemblage.      *
*                status = barre de statut à actualiser.                       *
*                id     = identifiant du groupe de progression à l'affichage. *
*                                                                             *
*  Description : S'assure que l'ensemble des aires est entièrement décodé.    *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

void ensure_all_mem_areas_are_filled(mem_area *list, size_t count, GProcContext *ctx, GtkStatusStack *status, activity_id_t id)
{
    size_t i;                               /* Boucle de parcours          */

    for (i = 0; i < count; i++)
        fill_mem_area(&list[i], list, count, ctx, status, id);

}



/* ---------------------------------------------------------------------------------- */
/*                         MANIPULATIONS PARALLELES DES ZONES                         */
/* ---------------------------------------------------------------------------------- */


/* Indique le type défini pour les tâches de traitement des zones. */
G_DEFINE_TYPE(GAreaCollector, g_area_collector, G_TYPE_DELAYED_WORK);


/******************************************************************************
*                                                                             *
*  Paramètres  : klass = classe à initialiser.                                *
*                                                                             *
*  Description : Initialise la classe des manipulations parallèles de zones.  *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_area_collector_class_init(GAreaCollectorClass *klass)
{
    GObjectClass *object;                   /* Autre version de la classe  */
    GDelayedWorkClass *work;                /* Version en classe parente   */

    object = G_OBJECT_CLASS(klass);

    object->dispose = (GObjectFinalizeFunc/* ! */)g_area_collector_dispose;
    object->finalize = (GObjectFinalizeFunc)g_area_collector_finalize;

    work = G_DELAYED_WORK_CLASS(klass);

    work->run = (run_task_fc)g_area_collector_process;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : collector = instance à initialiser.                          *
*                                                                             *
*  Description : Initialise des manipulations parallèles de zones.            *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_area_collector_init(GAreaCollector *collector)
{

}


/******************************************************************************
*                                                                             *
*  Paramètres  : collector = instance d'objet GLib à traiter.                 *
*                                                                             *
*  Description : Supprime toutes les références externes.                     *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_area_collector_dispose(GAreaCollector *collector)
{
    if (collector->run == (run_task_fc)g_area_collector_do_compute)
        g_object_unref(G_OBJECT(collector->binary));

    G_OBJECT_CLASS(g_area_collector_parent_class)->dispose(G_OBJECT(collector));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : collector = instance d'objet GLib à traiter.                 *
*                                                                             *
*  Description : Procède à la libération totale de la mémoire.                *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_area_collector_finalize(GAreaCollector *collector)
{
    G_OBJECT_CLASS(g_area_collector_parent_class)->finalize(G_OBJECT(collector));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : collector = opérations à mener.                              *
*                status    = barre de statut à tenir informée.                *
*                                                                             *
*  Description : Assure un traitement particulier concernant les zones.       *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_area_collector_process(GAreaCollector *collector, GtkStatusStack *status)
{
    collector->run(G_DELAYED_WORK(collector), status);

}


/******************************************************************************
*                                                                             *
*  Paramètres  : id      = identifiant pour signaler la progression courante. *
*                binary  = binaire chargé à conserver dans les zones définies.*
*                first   = localisation du début de la portion à traiter.     *
*                last    = localisation de la fin de la portion à traiter.    *
*                closing = indique si la tâche doit terminer l'analyse.       *
*                                                                             *
*  Description : Crée une tâche de calcul des zones binaires à désassembler.  *
*                                                                             *
*  Retour      : Tâche créée.                                                 *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static GAreaCollector *g_area_collector_new_intro(activity_id_t id, GLoadedBinary *binary, phys_t first, phys_t last, bool closing)
{
    GAreaCollector *result;            /* Tâche à retourner           */

    result = g_object_new(G_TYPE_AREA_COLLECTOR, NULL);

    result->id = id;
    result->run = (run_task_fc)g_area_collector_do_compute;

    result->areas = NULL;

    result->acount = 0;

    result->binary = binary;
    g_object_ref(G_OBJECT(binary));

    result->first = first;
    result->last = last;

    result->closing = closing;

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : fetching = récupération à mener.                             *
*                status   = barre de statut à tenir informée.                 *
*                                                                             *
*  Description : Construit une liste bornée de zones contigües.               *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_area_collector_do_compute(GAreaCollector *collector, GtkStatusStack *status)
{
    mem_area **list;                        /* Liste de zones à constituer */
    size_t *count;                          /* Nombre d'éléments intégrés  */
    vmpa2t first;                           /* Point de départ             */
    vmpa2t last;                            /* Point d'arrivée             */
    GExeFormat *format;                     /* Format du binaire           */
    vmpa2t prev;                            /* Dernière bordure rencontrée */
    bool state;                             /* Bilan d'une conversion      */
    GBinSymbol **symbols;                   /* Symboles à représenter      */
    size_t sym_count;                       /* Qté de symboles présents    */
    bool has_sym_index;                     /* Détermine une validité      */
    size_t sym_index;                       /* Prochain symbole non traité */
    GBinPortion *portions;                  /* Couche première de portions */

    void populate_with_symbols(const vmpa2t *limit)
    {
        GBinSymbol *symbol;                 /* Symbole en cours d'analyse  */
        SymbolType type;                    /* Nature d'un symbole         */
        const mrange_t *range;              /* Couverture d'un symbole     */
        vmpa2t end;                         /* Adresse de fin du symbole   */

        for (; sym_index < sym_count; sym_index++)
        {
            symbol = symbols[sym_index];

            type = g_binary_symbol_get_target_type(symbol);

            /**
             * On ne garde que les symboles renvoyant directement une ou
             * plusieurs instructions, c'est à dire les symboles valides
             * pour un appel à g_binary_symbol_get_instruction().
             *
             * Les instructions des autres symboles sont obtenues et mises
             * en place durant la procédure de désassemblage.
             */

            if (type == STP_ROUTINE || type == STP_ENTRY_POINT || type == STP_CODE_LABEL)
                continue;

            range = g_binary_symbol_get_range(symbol);

            if (get_mrange_length(range) == 0)
                continue;

            if (cmp_vmpa(get_mrange_addr(range), limit) >= 0)
                break;

            compute_mrange_end_addr(range, &end);

            /**
             * Si un symbole est à cheval entre deux zones, tant pis pour lui !
             */

            if (cmp_vmpa(&end, limit) > 0)
                break;

            insert_extra_symbol_into_mem_areas(*list, *count, symbol);

        }

    }

    void fill_gap(vmpa2t *old, vmpa2t *new, bool exec)
    {
        phys_t diff;                        /* Espace entre bordures       */
        mem_area *area;                     /* Zone avec valeurs à éditer  */

        diff = compute_vmpa_diff(old, new);

        /**
         * S'il existe un écart entre la dernière bordure ajoutée et
         * l'extréminité de la portion courante, on le comble !
         */

        if (diff > 0)
        {
            /* Zone tampon à constituer */

            *list = (mem_area *)realloc(*list, ++(*count) * sizeof(mem_area));

            area = &(*list)[*count - 1];

            init_mem_area_from_addr(area, old, diff, collector->binary);
            area->is_exec = exec;

            /* Insertion des symboles existants */

            if (!has_sym_index)
            {
                int cmp_vmpa_with_symbol(const vmpa2t *a, const GBinSymbol **s)
                {
                    return g_binary_symbol_cmp_with_vmpa(*s, a);
                }

                bsearch_index(old, symbols, sym_count, sizeof(GBinSymbol *),
                              (__compar_fn_t)cmp_vmpa_with_symbol, &sym_index);

                has_sym_index = true;

            }

            populate_with_symbols(new);

            /* Avancée du curseur */

            copy_vmpa(old, new);

            gtk_status_stack_update_activity_value(status, collector->id, diff);

        }

        else
        {
            /**
             * La comparaison entre les bordures se réalise selon les positions
             * physiques renseignées.
             *
             * Aussi, même dans le cas d'une jointure sans espace, il se peut que
             * la transition concerne deux zones aux adresses virtuelles non
             * consécutives.
             *
             * Comme "old" est mise à jour pour devenir le point de départ de
             * la zone suivante, on se doit de même à jour les deux positions :
             * physique et virtuelle.
             */

            copy_vmpa(old, new);

        }

    }

    bool build_area_from_portion(GBinPortion *portion, GBinPortion *parent, BinaryPortionVisit visit, void *unused)
    {
        const mrange_t *range;              /* Espace de portion à traiter */
        vmpa2t border;                      /* Nouvelle bordure rencontrée */
        bool on_track;                      /* Le tronçon courant est bon ?*/
        PortionAccessRights rights;         /* Droits d'accès à analyser   */

        range = g_binary_portion_get_range(portion);

        if (visit == BPV_ENTER)
        {
            copy_vmpa(&border, get_mrange_addr(range));

            on_track = cmp_vmpa(&first, &border) <= 0 && cmp_vmpa(&border, &last) < 0;

            if (on_track)
            {
                rights = (parent != NULL ? g_binary_portion_get_rights(parent) : PAC_NONE);
                fill_gap(&prev, &border, rights & PAC_EXEC);
            }
            else
                copy_vmpa(&prev, &border);

        }

        else if (visit == BPV_SHOW)
        {
            copy_vmpa(&border, get_mrange_addr(range));

            on_track = cmp_vmpa(&first, &border) <= 0 && cmp_vmpa(&border, &last) < 0;

            if (on_track)
            {
                rights = (parent != NULL ? g_binary_portion_get_rights(parent) : PAC_NONE);
                fill_gap(&prev, &border, rights & PAC_EXEC);

                compute_mrange_end_addr(range, &border);

                rights = g_binary_portion_get_rights(portion);
                fill_gap(&prev, &border, rights & PAC_EXEC);

            }
            else
                compute_mrange_end_addr(range, &prev);

        }

        else if (visit == BPV_EXIT)
        {
            compute_mrange_end_addr(range, &border);

            if (collector->closing)
                on_track = cmp_vmpa(&first, &border) <= 0 && cmp_vmpa(&border, &last) <= 0;
            else
                on_track = cmp_vmpa(&first, &border) <= 0 && cmp_vmpa(&border, &last) < 0;

            if (on_track)
            {
                rights = (parent != NULL ? g_binary_portion_get_rights(parent) : PAC_NONE);
                fill_gap(&prev, &border, rights & PAC_EXEC);
            }
            else
                copy_vmpa(&prev, &border);

        }

#ifndef NDEBUG
        else
            assert(false);
#endif

        return (cmp_vmpa(&prev, &last) < 0);

    }

    list = &collector->areas;
    count = &collector->acount;

    init_vmpa(&first, collector->first, VMPA_NO_VIRTUAL);
    init_vmpa(&last, collector->last, VMPA_NO_VIRTUAL);

    format = g_loaded_binary_get_format(collector->binary);

#ifndef NDEBUG
    state = g_exe_format_translate_offset_into_vmpa(format, 0, &prev);
    assert(state);
#else
    g_exe_format_translate_offset_into_vmpa(format, 0, &prev);
#endif

    symbols = g_binary_format_get_symbols(G_BIN_FORMAT(format), &sym_count);

    has_sym_index = false;

    portions = g_exe_format_get_portions(format);

    g_binary_portion_visit(portions, (visit_portion_fc)build_area_from_portion, NULL);

    g_object_unref(G_OBJECT(portions));

    g_object_unref(G_OBJECT(format));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : gid    = groupe de travail impliqué.                         *
*                status = barre de statut à tenir informée.                   *
*                binary = binaire analysé contenant quantités d'infos.        *
*                length = quantité d'octets à traiter au total.               *
*                count  = nombre de zones mises en place. [OUT]               *
*                                                                             *
*  Description : Détermine une liste de zones contigües à traiter.            *
*                                                                             *
*  Retour      : Liste de zones mémoire à libérer après usage.                *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

mem_area *collect_memory_areas(wgroup_id_t gid, GtkStatusStack *status, GLoadedBinary *binary, phys_t length, size_t *count)
{
    mem_area *result;                       /* Liste finale à retourner    */
    guint runs_count;                       /* Qté d'exécutions parallèles */
    GAreaCollector **collectors;            /* Collecteurs à suivre        */
    phys_t run_size;                        /* Volume réparti par exécution*/
    GWorkQueue *queue;                      /* Gestionnaire de différés    */
    activity_id_t id;                       /* Identifiant de progression  */
    guint i;                                /* Boucle de parcours          */
    phys_t first;                           /* Début de zone de traitement */
    bool closing;                           /* Détection de fin en amont   */
    phys_t last;                            /* Fin de zone de traitement   */

    runs_count = g_get_num_processors();

    collectors = (GAreaCollector **)calloc(runs_count, sizeof(GAreaCollector *));

    run_size = length / runs_count;

    queue = get_work_queue();

    id = gtk_status_stack_add_activity(status, _("Computing memory areas to disassemble"), length);

    for (i = 0; i < runs_count; i++)
    {
        first = i * run_size;

        closing = ((i + 1) == runs_count);

        if (closing)
            last = length;
        else
            last = first + run_size;

        collectors[i] = g_area_collector_new_intro(id, binary, first, last, closing);

        g_object_ref(G_OBJECT(collectors[i]));
        g_work_queue_schedule_work(queue, G_DELAYED_WORK(collectors[i]), gid);

    }

    g_work_queue_wait_for_completion(queue, gid);

    /* Récupération des aires */

    result = NULL;
    *count = 0;

    for (i = 0; i < runs_count; i++)
    {
        result = (mem_area *)realloc(result, (*count + collectors[i]->acount) * sizeof(mem_area));

        memcpy(&result[*count], collectors[i]->areas, collectors[i]->acount * sizeof(mem_area));
        *count += collectors[i]->acount;

        g_object_unref(G_OBJECT(collectors[i]));

    }

    /* Fin */

    free(collectors);

    gtk_status_stack_remove_activity(status, id);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : id    = identifiant pour signaler la progression courante.   *
*                list  = liste des zones en place à parcourir.                *
*                begin = indice de la première zone à traiter.                *
*                end   = indice de la première zone à ne pas traiter.         *
*                                                                             *
*  Description : Crée une tâche de récupération d'instructions différée.      *
*                                                                             *
*  Retour      : Tâche créée.                                                 *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static GAreaCollector *g_area_collector_new_outro(activity_id_t id, mem_area *list, size_t begin, size_t end)
{
    GAreaCollector *result;            /* Tâche à retourner           */

    result = g_object_new(G_TYPE_AREA_COLLECTOR, NULL);

    result->id = id;
    result->run = (run_task_fc)g_area_collector_do_collect;

    result->areas = list;

    result->begin = begin;
    result->end = end;

    result->collected = NULL;
    result->ccount = 0;

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : fetching = récupération à mener.                             *
*                status   = barre de statut à tenir informée.                 *
*                                                                             *
*  Description : Assure la récupération d'instructions en différé.            *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_area_collector_do_collect(GAreaCollector *collector, GtkStatusStack *status)
{
    size_t i;                               /* Boucle de parcours          */

    for (i = collector->begin; i < collector->end; i++)
    {
        collector->collected = get_instructions_from_mem_area(&collector->areas[i],
                                                              collector->collected, &collector->ccount);

        fini_mem_area(&collector->areas[i]);

        gtk_status_stack_update_activity_value(status, collector->id, 1);

    }

}


/******************************************************************************
*                                                                             *
*  Paramètres  : gid    = groupe de travail impliqué.                         *
*                status = barre de statut à tenir informée.                   *
*                list   = liste des zones de données à relire puis libérer.   *
*                acount = taille de cette liste de zones.                     *
*                icount = nombre d'instructions récupérées. [OUT]             *
*                                                                             *
*  Description : Rassemble les instructions conservées dans des zones données.*
*                                                                             *
*  Retour      : Liste d'instructions rassemblées.                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

GArchInstruction **collect_disassembled_instructions(wgroup_id_t gid, GtkStatusStack *status, mem_area *list, size_t acount, size_t *icount)
{
    GArchInstruction **result;              /* Liste finale à retourner    */
    guint runs_count;                       /* Qté d'exécutions parallèles */
    GAreaCollector **collectors;            /* Collecteurs à suivre        */
    size_t run_size;                        /* Volume réparti par exécution*/
    GWorkQueue *queue;                      /* Gestionnaire de différés    */
    activity_id_t id;                       /* Identifiant de progression  */
    guint i;                                /* Boucle de parcours          */
    size_t begin;                           /* Début de bloc de traitement */
    size_t end;                             /* Fin d'un bloc de traitement */

    runs_count = g_get_num_processors();

    collectors = (GAreaCollector **)calloc(runs_count, sizeof(GAreaCollector *));

    run_size = acount / runs_count;

    queue = get_work_queue();

    id = gtk_status_stack_add_activity(status, _("Collecting all disassembled instructions"), acount);

    for (i = 0; i < runs_count; i++)
    {
        begin = i * run_size;

        if ((i + 1) == runs_count)
            end = acount;
        else
            end = begin + run_size;

        collectors[i] = g_area_collector_new_outro(id, list, begin, end);

        g_object_ref(G_OBJECT(collectors[i]));
        g_work_queue_schedule_work(queue, G_DELAYED_WORK(collectors[i]), gid);

    }

    g_work_queue_wait_for_completion(queue, gid);

    /* Récupération des instructions */

    result = NULL;
    *icount = 0;

    for (i = 0; i < runs_count; i++)
    {
        result = (GArchInstruction **)realloc(result,
                                              (*icount + collectors[i]->ccount) * sizeof(GArchInstruction *));

        memcpy(&result[*icount], collectors[i]->collected, collectors[i]->ccount * sizeof(GArchInstruction *));
        *icount += collectors[i]->ccount;

        g_object_unref(G_OBJECT(collectors[i]));

    }

    /* Fin */

    free(collectors);

    free(list);

    gtk_status_stack_remove_activity(status, id);

    return result;

}
